/*
 * * Copyright (c) 2018, Fabien Feschet <fabien.feschet@dg-medical.fr>
 * All rights reserved.
 *
 * This copyright applies for functions:
 *   symmetrize, load_balancing_2d
 *
 * Licence   : Simplified BSD License see LICENSE file.
 *
 * **********************************************************************
 * Copyright (c) 2011, Marc Lebrun <marc.lebrun@cmla.ens-cachan.fr>
 * All rights reserved.
 *
 * This copyright applies for functions:
 *   load_image, save_image, add_noise, compute_psnr,
 *   color_space_transform, estimate_sigma
 *
 * This program is free software: you can use, modify and/or
 * redistribute it under the terms of the GNU General Public
 * License as published by the Free Software Foundation, either
 * version 3 of the License, or (at your option) any later
 * version. You should have received a copy of this license along
 * this program. If not, see <http://www.gnu.org/licenses/>.
 */

/**
 * @file utilities.cpp
 * @brief Utilities functions
 *
 * @author Fabien Feschet <fabien.feschet@dg-medical.fr>
 *     and Marc Lebrun <marc.lebrun@cmla.ens-cachan.fr>
 **/

#include <iostream>
#include <cstdlib>
#include <ctime>
#include <unistd.h>
#include <cmath>
#include <iterator>

extern "C" {
#include "./iio/iio.h"
}

#include "mt19937ar.h"
#include "io_png.h"
#include "hosvd2.h"

 using namespace std;

/**
 * @brief Load image via IIO, check the number of channels
 *
 * @param name : name of the image to read
 * @param img : vector which will contain the image : R, G and B concatenated
 * @param width, height, chnls : size of the image
 *
 * @return EXIT_SUCCESS if the image has been loaded, EXIT_FAILURE otherwise
 **/
int load_image_IIO(
   const char* name
,   vector<float> &img
,   unsigned * width
,   unsigned * height
,   unsigned * chnls
){
   //! read input image
   cout << endl << "Read input image...";
   int h, w, c;
   float *tmp = NULL;
   tmp = iio_read_image_float_split(name, &w, &h, &c);
   if (!tmp)
   {
       cout << "error :: " << name << " not found or not a correct png or tiff input image" << endl;
       return EXIT_FAILURE;
   }
   cout << "done." << endl;
   //! test if image is really a color image and exclude the alpha channel
   if (c > 2)
   {
       unsigned k = 0;
       while (k < w * h && tmp[k] == tmp[w * h + k] && tmp[k] == tmp[2 * w * h + k]) {
           k++;
       }
       c = (k == w * h ? 1 : 3);
   }
   //! Some image informations
   cout << "image size :" << endl;
   cout << " - width          = " << w << endl;
   cout << " - height         = " << h << endl;
   cout << " - nb of channels = " << c << endl;
   //! Initializations
   *width  = static_cast<unsigned>(w);
   *height = static_cast<unsigned>(h);
   *chnls  = static_cast<unsigned>(c);
   img.resize(w * h * c);
   for (unsigned k = 0; k < w * h * c; k++)
       img[k] = tmp[k];
   return EXIT_SUCCESS;
}

int load_image_IIO(
   const char* name
,   vector<float> &img
,   unsigned width
,   unsigned height
,   unsigned chnls
){
   //! read input image
   cout << endl << "Read input image...";
   int h, w, c;
   float *tmp = NULL;
   tmp = iio_read_image_float_split(name, &w, &h, &c);
   if (!tmp)
   {
       cout << "error :: " << name << " not found or not a correct png or tiff input image" << endl;
       return EXIT_FAILURE;
   }
   cout << "done." << endl;
    //! check size
    if ((width != w) || (height != h) || (chnls != c))
    {
        cout << "error :: size does not match." << endl;
        return EXIT_FAILURE; 
    }
   //! Some image informations
   cout << "image size :" << endl;
   cout << " - width          = " << w << endl;
   cout << " - height         = " << h << endl;
   cout << " - nb of channels = " << c << endl;
   //! Initializations
   img.resize(w * h * c);
   for (unsigned k = 0; k < w * h * c; k++)
       img[k] = tmp[k];
   return EXIT_SUCCESS;
}

 /**
  * @brief Load image, check the number of channels
  *
  * @param name : name of the image to read
  * @param img : vector which will contain the image : R, G and B concatenated
  * @param width, height, chnls : size of the image
  *
  * @return EXIT_SUCCESS if the image has been loaded, EXIT_FAILURE otherwise
  **/
int load_image(
    const char* name
,   vector<float> &img
,   unsigned * width
,   unsigned * height
,   unsigned * chnls
){
    //! read input image
    cout << endl << "Read input image...";
    size_t h, w, c;
    float *tmp = NULL;
    tmp = read_png_f32(name, &w, &h, &c);
    if (!tmp)
    {
        cout << "error :: " << name << " not found or not a correct png image" << endl;
        return EXIT_FAILURE;
    }
    cout << "done." << endl;
    //! test if image is really a color image and exclude the alpha channel
    if (c > 2)
    {
        unsigned k = 0;
        while (k < w * h && tmp[k] == tmp[w * h + k] && tmp[k] == tmp[2 * w * h + k]) {
            k++;
        }
        c = (k == w * h ? 1 : 3);
    }
    //! Some image informations
    cout << "image size :" << endl;
    cout << " - width          = " << w << endl;
    cout << " - height         = " << h << endl;
    cout << " - nb of channels = " << c << endl;
    //! Initializations
    *width  = w;
    *height = h;
    *chnls  = c;
    img.resize(w * h * c);
    for (unsigned k = 0; k < w * h * c; k++)
        img[k] = tmp[k];
    return EXIT_SUCCESS;
}

/**
  * @brief Load image, check the number of channels
  *
  * @param name : name of the image to read
  * @param img : vector which will contain the image : R, G and B concatenated
  * @param width, height, chnls : size of the image
  *
  * @return EXIT_SUCCESS if the image has been loaded, EXIT_FAILURE otherwise
  **/
int load_image(
    const char* name
,   vector<float> &img
,   unsigned  width
,   unsigned  height
,   unsigned  chnls
){
    //! read input image
    cout << endl << "Read input image...";
    size_t h, w, c;
    float *tmp = NULL;
    tmp = read_png_f32(name, &w, &h, &c);
    if (!tmp)
    {
        cout << "error :: " << name << " not found or not a correct png image" << endl;
        return EXIT_FAILURE;
    }
    cout << "done." << endl;
    //! check size
    if ((width != w) || (height != h) || (chnls != c))
    {
        cout << "error :: size does not match." << endl;
        return EXIT_FAILURE; 
    }
    //! Some image informations
    cout << "image size :" << endl;
    cout << " - width          = " << w << endl;
    cout << " - height         = " << h << endl;
    cout << " - nb of channels = " << c << endl;
    //! Initializations
    img.resize(w * h * c);
    for (unsigned k = 0; k < w * h * c; k++)
        img[k] = tmp[k];
    return EXIT_SUCCESS;
}


/**
 * @brief write image
 *
 * @param name : path+name+extension of the image
 * @param img : vector which contains the image
 * @param width, height, chnls : size of the image
 *
 * @return EXIT_SUCCESS if the image has been saved, EXIT_FAILURE otherwise
 **/
int save_image(
    char* name
,   std::vector<float> &img
,   const unsigned width
,   const unsigned height
,   const unsigned chnls
){
    //! Allocate Memory
    float* tmp = new float[width * height * chnls];

    //! Check for boundary problems
    for (unsigned k = 0; k < width * height * chnls; k++)
        tmp[k] = (img[k] > 255.0f ? 255.0f : (img[k] < 0.0f ? 0.0f : img[k]));

    if (write_png_f32(name, tmp, width, height, chnls) != 0)
    {
        cout << "... failed to save png image " << name << endl;
        return EXIT_FAILURE;
    }

    //! Free Memory
    delete[] tmp;

    return EXIT_SUCCESS;
}

/**
 * @brief write image.
 *
 * @param name : path+name+extension of the image
 * @param img : vector which contains the image
 * @param width, height, chnls : size of the image
 *
 * @return EXIT_SUCCESS if the image has been saved, EXIT_FAILURE otherwise
 **/
int save_image_IIO(
        char* name
    ,   std::vector<float> &img
    ,   const unsigned width
    ,   const unsigned height
    ,   const unsigned chnls
    ){
        //! Allocate Memory
        float* tmp = new float[width * height * chnls];

        //! Check for boundary problems
        for (unsigned k = 0; k < width * height * chnls; k++)
            tmp[k] = (img[k] > 255.0f ? 255.0f : (img[k] < 0.0f ? 0.0f : img[k]));

        iio_write_image_float_split(name, tmp,
                                    width,
                                    height,
                                    chnls);

        //! Free Memory
        delete[] tmp;

        return EXIT_SUCCESS;
    }

/**
 * @brief add noise to img
 *
 * @param img : original noise-free image
 * @param img_noisy = img + noise
 * @param sigma : standard deviation of the noise
 *
 * @return none.
 **/
void add_noise(
    const vector<float> &img
,   vector<float> &img_noisy
,   const float sigma
){
    const unsigned size = img.size();
    if (img_noisy.size() != size)
        img_noisy.resize(size);

    mt_init_genrand((unsigned long int) time (NULL) + (unsigned long int) getpid());

    for (unsigned k = 0; k < size; k++)
    {
        double a = mt_genrand_res53();
        double b = mt_genrand_res53();
        double z = (double)(sigma) * sqrt(-2.0 * log(a)) * cos(2.0 * M_PI * b);

        img_noisy[k] =  img[k] + (float) z;
    }
}

/**
 * @brief Add boundaries by symmetry
 *
 * @param img : image to symmetrize
 * @param img_sym : will contain img with symetrized boundaries
 * @param width, height, chnls : size of img
 * @param N : size of the boundary (square case for extension)
 *
 * @return none.
 **/
void symmetrize(
    const std::vector<float> &img
,   std::vector<float> &img_sym
,   const unsigned width
,   const unsigned height
,   const unsigned chnls
,   const unsigned N
){
  symmetrize(img,img_sym,width,height,chnls,N,N);
}

/**
 * @brief Add boundaries by symmetry
 *
 * @param img : image to symmetrize
 * @param img_sym : will contain img with symetrized boundaries
 * @param width, height, chnls : size of img
 * @param P x Q : size of the boundary (rectangular case)
 *
 * @return none.
 **/
void symmetrize(
    const std::vector<float> &img
,   std::vector<float> &img_sym
,   const unsigned width
,   const unsigned height
,   const unsigned chnls
,   const unsigned P
,   const unsigned Q
){
    //! Declaration
    const unsigned w = width + 2 * P;
    const unsigned h = height + 2 * Q;

    if (img_sym.size() != w * h * chnls)
        img_sym.resize(w * h * chnls);

    std::vector<float>::const_iterator dc_in  = img.cbegin();
    std::vector<float>::iterator dc_out = img_sym.begin();

    for (unsigned c = 0; c < chnls; c++)
    {
        auto dc_inb  = dc_in; // backup
        auto dc_outb = dc_out;// backup
        dc_out = dc_out + Q * w + P;

        //! Center of the image
        for (unsigned i = 0; i < height; i++) {
            // STL copy
            std::copy_n(dc_in , width , dc_out);
            // Move iterators
            dc_in  += width;
            dc_out += w;
        }

        //! Top and bottom
        auto dc_out_t  = dc_outb + Q*w + P; // first point of the center image
        auto dc_out_tb = dc_out_t - w;      // first point of top

        auto dc_out_bb= dc_outb + (Q+height)*w + P; // first point of bottom
        auto dc_out_b = dc_out_bb - w;              // first point center image bottom

        for (unsigned j = 0; j < Q; j++) {
            // STL copy: top then bottom
            std::copy_n(dc_out_t  , width, dc_out_tb);
            std::copy_n(dc_out_b  , width, dc_out_bb);
            // Move top then bottom
            dc_out_t += w; dc_out_tb -= w;
            dc_out_b -= w; dc_out_bb += w;
        }

        //! Left and Right
        auto dc_out_l  = dc_outb + P; // left in
        //! This was c++14
        //auto dc_out_lb = std::make_reverse_iterator(dc_out_l); // left out
        std::reverse_iterator< std::vector<float>::iterator > dc_out_lb ( dc_out_l ); // left out

        auto dc_out_rb = dc_outb + w - P; // right out
        //! This was c++14
        //auto dc_out_r  = std::make_reverse_iterator(dc_out_rb); // right in
        std::reverse_iterator< std::vector<float>::iterator > dc_out_r ( dc_out_rb ); // right in

        for (unsigned j = 0; j < h; j++) {
            // STL copy left then right
            std::copy_n(dc_out_l, P, dc_out_lb);
            std::copy_n(dc_out_r, P, dc_out_rb);
            // Move left then right
            dc_out_l += w; dc_out_lb -=w;
            dc_out_r -= w; dc_out_rb +=w;
        }

        // Move iterators to next chanel
        dc_in  = dc_inb + width*height;
        dc_out = dc_outb + w*h;
    }
}

/**
 * @brief Add boundaries by symmetry
 *
 * @param img : image to symmetrize
 * @param img_sym : will contain img with symetrized boundaries
 * @param width, height, chnls : size of img
 * @param P x Q : size of the boundary (rectangular case)
 *
 * @return none.
 **/
void symmetrize(
    const std::vector<float> &img
,   std::vector<float> &img_sym
,   const unsigned width
,   const unsigned height
,   const unsigned chnls
,   const unsigned P
,   const unsigned Q
,   const Tiling &region
){
    // Get region of support
    unsigned xmin, xmax, ymin, ymax;
    std::tie(xmin, xmax, ymin, ymax) = region;

    //! Declaration of output size
    const unsigned w = xmax - xmin + 2 * P;
    const unsigned h = ymax - ymin + 2 * Q;

    if (img_sym.size() != w * h * chnls)
        img_sym.resize(w * h * chnls);

    auto dc_in  = img.begin()+ymin*width+xmin;
    auto dc_out = img_sym.begin();

    for (unsigned c = 0; c < chnls; c++)
    {
        auto dc_inb  = dc_in; // backup
        auto dc_outb = dc_out;// backup
        dc_out = dc_out + Q * w + P;

        //! Center of the image
        for (unsigned i = ymin; i < ymax; i++) {
            // STL copy
            std::copy_n(dc_in , width , dc_out);
            // Move iterators
            dc_in  += width;
            dc_out += w;
        }

        //! Starting input image  is not used anymore
        //! Top and bottom
        auto dc_out_t  = dc_outb + Q*w + P; // first point of the center image
        auto dc_out_tb = dc_out_t - w;      // first point of top

        auto dc_out_bb= dc_outb + (Q+h)*w + P; // first point of bottom
        auto dc_out_b = dc_out_bb - w;              // first point center image bottom

        for (unsigned j = 0; j < Q; j++) {
            // STL copy: top then bottom
            std::copy_n(dc_out_t , w, dc_out_tb);
            std::copy_n(dc_out_b , w, dc_out_bb);
            // Move top then bottom
            dc_out_t += w; dc_out_tb -= w;
            dc_out_b -= w; dc_out_bb += w;
        }

        //! Left and Right
        auto dc_out_l  = dc_outb + P; // left in
        // auto dc_out_lb = std::make_reverse_iterator(dc_out_l); // left out c++14 only
        std::reverse_iterator< std::vector<float>::iterator > dc_out_lb (dc_out_l);

        auto dc_out_rb = dc_outb + w - P; // right out
        //auto dc_out_r  = std::make_reverse_iterator(dc_out_rb); // right in c++14 only
        std::reverse_iterator< std::vector<float>::iterator > dc_out_r( dc_out_rb);

        for (unsigned j = 0; j < h; j++) {
            // STL copy left then right
            std::copy_n(dc_out_l, P, dc_out_lb);
            std::copy_n(dc_out_r, P, dc_out_rb);
            // Move left then right
            dc_out_l += w; dc_out_lb -=w;
            dc_out_r -= w; dc_out_rb +=w;
        }

        // Move iterators to next chanel
        dc_in  = dc_inb + width*height;
        dc_out = dc_outb + w*h;
    }
}

//! Best load_balance decomposition from image size and
//! number of processors.
//! Condition: a decompositoin must contain at least one PxQ block
int load_balancing_2d(unsigned int nb_threads,
        unsigned int width,
        unsigned int height,
        Tiling_2d &decomposition)
{
    auto divisors = divisors_table[nb_threads];
    float max_ratio = 0.0f;
    unsigned int idx_ratio = 0;
    unsigned int order = 0; // 1 is width x height and 2 is height x width
    for (auto d = divisors.begin() ; d != divisors.end() ; ++d) {
        unsigned int u = d->first; unsigned int v = d->second;
        // Compute load balancing using n = u x v
        //! Case u x v in width x height
        unsigned int alpha_w = width / u;
        unsigned int alpha_h = height / v;
        // Since width  = (1+alpha_w)u' + alpha_w(u-u')
        // and   height = (1+alpha_h)v' + alpha_h(v-v')
        // the smallest block is of size alpha_w x alpha_h
        // and the largest one is (1+alpha_w) x (1+alpha_h)
        float ratio = static_cast<float>(alpha_w * alpha_h) / static_cast<float> ((1+alpha_w)*(1+alpha_h));
        if ((ratio > max_ratio)) {
            idx_ratio = (d-divisors.begin());
            max_ratio = ratio;
            order = 1;
        }
        //! Case v x u in width x height
        alpha_w = width / v;
        alpha_h = height / u;
        // Ratio
        ratio = static_cast<float>(alpha_w * alpha_h) / static_cast<float> ((1+alpha_w)*(1+alpha_h));
        if ((ratio > max_ratio)) {
            idx_ratio = (d-divisors.begin());
            max_ratio = ratio;
            order = 2;
        }
    }
    // Create decomposition X and Y decomposition
    std::pair< unsigned int, unsigned int > uv = divisors[idx_ratio];
    unsigned int u = (order == 1 ? uv.first : uv.second);
    unsigned int v = (order == 1 ? uv.second : uv.first);
    std::vector< unsigned int > boundary_X(u+1);
    std::vector< unsigned int > boundary_Y(v+1);
    unsigned int alpha_w = width / u;
    unsigned int u_prime = width % u;
    boundary_X[0] = 0;
    for (auto k = 1u ; k <= u_prime ; ++k)
        boundary_X[k] = boundary_X[k-1] + (1+alpha_w);
    for (auto k = 1u ; k <= u - u_prime ; ++k)
        boundary_X[u_prime+k] = boundary_X[u_prime+k-1] + alpha_w;
    unsigned int alpha_h = height / v;
    unsigned int v_prime = height % v;
    boundary_Y[0] = 0;
    for (auto k = 1u ; k <= v_prime ; ++k)
        boundary_Y[k] = boundary_Y[k-1] + (1+alpha_h);
    for (auto k = 1u ; k <= v - v_prime ; ++k)
        boundary_Y[v_prime+k] = boundary_Y[v_prime+k-1] + alpha_h;
    // Return result
    decomposition.resize(nb_threads);
    auto dcp = decomposition.begin();
    for (auto kv = 0u ; kv < v ; ++kv)
        for (auto ku = 0u ; ku < u ; ++ku) {
            *dcp = std::move(std::make_tuple(boundary_X[ku], boundary_X[ku+1],boundary_Y[kv],boundary_Y[kv+1]));
            ++dcp;
        }
    return EXIT_SUCCESS;
}

/**
 * @brief Compute PSNR and RMSE between img_1 and img_2
 *
 * @param img_1 : pointer to an allocated array of pixels.
 * @param img_2 : pointer to an allocated array of pixels.
 * @param psnr  : will contain the PSNR
 * @param rmse  : will contain the RMSE
 *
 * @return EXIT_FAILURE if both images haven't the same size.
 **/
int compute_psnr(
    const vector<float> &img_1
,   const vector<float> &img_2
,   float *psnr
,   float *rmse
)
{
    if (img_1.size() != img_2.size())
    {
        cout << "Can't compute PSNR & RMSE: images have different sizes: " << endl;
        cout << "img_1 : " << img_1.size() << endl;
        cout << "img_2 : " << img_2.size() << endl;
        return EXIT_FAILURE;
    }

    float tmp = 0.0f;
    for (unsigned k = 0; k < img_1.size(); k++)
        tmp += (img_1[k] - img_2[k]) * (img_1[k] - img_2[k]);

    (*rmse) = sqrtf(tmp / (float) img_1.size());
    (*psnr) = 20.0f * log10f(255.0f / (*rmse));

    return EXIT_SUCCESS;
}


/**
 * @brief Transform the color space of the image
 *
 * @param img: image to transform
 * @param color_space: choice between OPP, YUV, YCbCr, RGB
 * @param width, height, chnls: size of img
 * @param rgb2yuv: if true, transform the color space
 *        from RGB to YUV, otherwise do the inverse
 *
 * @return EXIT_FAILURE if color_space has not expected
 *         type, otherwise return EXIT_SUCCESS.
 **/
int color_space_transform(
    vector<float> &img
,   const unsigned color_space
,   const unsigned width
,   const unsigned height
,   const unsigned chnls
,   const bool rgb2yuv
){
    if (chnls == 1 || color_space == RGB)
        return EXIT_SUCCESS;

    //! Declarations
    vector<float> tmp;
    tmp.resize(chnls * width * height);
    const unsigned red   = 0;
    const unsigned green = width * height;
    const unsigned blue  = width * height * 2;

    //! Transformations depending on the mode
    if (color_space == YUV)
    {
        if (rgb2yuv)
        {
        #pragma omp parallel for
            for (unsigned k = 0; k < width * height; k++)
            {
                //! Y
                tmp[k + red  ] =  0.299f   * img[k + red] + 0.587f   * img[k + green] + 0.114f   * img[k + blue];
                //! U
                tmp[k + green] = -0.14713f * img[k + red] - 0.28886f * img[k + green] + 0.436f   * img[k + blue];
                //! V
                tmp[k + blue ] =  0.615f   * img[k + red] - 0.51498f * img[k + green] - 0.10001f * img[k + blue];
            }
        }
        else
        {
        #pragma omp parallel for
            for (unsigned k = 0; k < width * height; k++)
            {
                //! Red   channel
                tmp[k + red  ] = img[k + red] + 1.13983f * img[k + blue];
                //! Green channel
                tmp[k + green] = img[k + red] - 0.39465f * img[k + green] - 0.5806f * img[k + blue];
                //! Blue  channel
                tmp[k + blue ] = img[k + red] + 2.03211f * img[k + green];
            }
        }
    }
    else if (color_space == YCBCR)
    {
        if (rgb2yuv)
        {
        #pragma omp parallel for
            for (unsigned k = 0; k < width * height; k++)
            {
                //! Y
                tmp[k + red  ] =  0.299f * img[k + red] + 0.587f * img[k + green] + 0.114f * img[k + blue];
                //! U
                tmp[k + green] = -0.169f * img[k + red] - 0.331f * img[k + green] + 0.500f * img[k + blue];
                //! V
                tmp[k + blue ] =  0.500f * img[k + red] - 0.419f * img[k + green] - 0.081f * img[k + blue];
            }
        }
        else
        {
        #pragma omp parallel for
            for (unsigned k = 0; k < width * height; k++)
            {
                //! Red   channel
                tmp[k + red  ] = 1.000f * img[k + red] + 0.000f * img[k + green] + 1.402f * img[k + blue];
                //! Green channel
                tmp[k + green] = 1.000f * img[k + red] - 0.344f * img[k + green] - 0.714f * img[k + blue];
                //! Blue  channel
                tmp[k + blue ] = 1.000f * img[k + red] + 1.772f * img[k + green] + 0.000f * img[k + blue];
            }
        }
    }
    else if (color_space == OPP)
    {
        if (rgb2yuv)
        {
        #pragma omp parallel for
            for (unsigned k = 0; k < width * height; k++)
            {
                //! Y
                tmp[k + red  ] =  0.333f * img[k + red] + 0.333f * img[k + green] + 0.333f * img[k + blue];
                //! U
                tmp[k + green] =  0.500f * img[k + red] + 0.000f * img[k + green] - 0.500f * img[k + blue];
                //! V
                tmp[k + blue ] =  0.250f * img[k + red] - 0.500f * img[k + green] + 0.250f * img[k + blue];
            }
        }
        else
        {
        #pragma omp parallel for
            for (unsigned k = 0; k < width * height; k++)
            {
                //! Red   channel
                tmp[k + red  ] = 1.0f * img[k + red] + 1.0f * img[k + green] + 0.666f * img[k + blue];
                //! Green cha
                tmp[k + green] = 1.0f * img[k + red] + 0.0f * img[k + green] - 1.333f * img[k + blue];
                //! Blue  cha
                tmp[k + blue ] = 1.0f * img[k + red] - 1.0f * img[k + green] + 0.666f * img[k + blue];
            }
        }
    }
    else
    {
        cout << "Wrong type of transform. Must be OPP, YUV, or YCbCr!!" << endl;
        return EXIT_FAILURE;
    }

    #pragma omp parallel for
        for (unsigned k = 0; k < width * height * chnls; k++)
            img[k] = tmp[k];

    return EXIT_SUCCESS;
}

/**
 * @brief Estimate sigma on each channel according to
 *        the choice of the color_space.
 *
 * @param sigma: estimated standard deviation of the noise;
 * @param sigma_Y : noise on the first channel;
 * @param sigma_U : (if chnls > 1) noise on the second channel;
 * @param sigma_V : (if chnls > 1) noise on the third channel;
 * @param chnls : number of channels of the image;
 * @param color_space : choice between OPP, YUV, YCbCr. If not
 *        then we assume that we're still in RGB space.
 *
 * @return EXIT_FAILURE if color_space has not expected
 *         type, otherwise return EXIT_SUCCESS.
 **/
int estimate_sigma(
    const float sigma
,   std::vector<float> &sigma_table
,   const unsigned chnls
,   const unsigned color_space
){
    if (chnls == 1)
        sigma_table[0] = sigma;
    else
    {
        if (color_space == YUV)
        {
            //! Y
            sigma_table[0] = sqrtf(0.299f * 0.299f + 0.587f * 0.587f + 0.114f * 0.114f) * sigma;
            //! U
            sigma_table[1] = sqrtf(0.14713f * 0.14713f + 0.28886f * 0.28886f + 0.436f * 0.436f) * sigma;
            //! V
            sigma_table[2] = sqrtf(0.615f * 0.615f + 0.51498f * 0.51498f + 0.10001f * 0.10001f) * sigma;
        }
        else if (color_space == YCBCR)
        {
            //! Y
            sigma_table[0] = sqrtf(0.299f * 0.299f + 0.587f * 0.587f + 0.114f * 0.114f) * sigma;
            //! U
            sigma_table[1] = sqrtf(0.169f * 0.169f + 0.331f * 0.331f + 0.500f * 0.500f) * sigma;
            //! V
            sigma_table[2] = sqrtf(0.500f * 0.500f + 0.419f * 0.419f + 0.081f * 0.081f) * sigma;
        }
        else if (color_space == OPP)
        {
            //! Y
            sigma_table[0] = sqrtf(0.333f * 0.333f + 0.333f * 0.333f + 0.333f * 0.333f) * sigma;
            //! U
            sigma_table[1] = sqrtf(0.5f * 0.5f + 0.0f * 0.0f + 0.5f * 0.5f) * sigma;
            //! V
            sigma_table[2] = sqrtf(0.25f * 0.25f + 0.5f * 0.5f + 0.25f * 0.25f) * sigma;
        }
        else if (color_space == RGB)
        {
            //! Y
            sigma_table[0] = sigma;
            //! U
            sigma_table[1] = sigma;
            //! V
            sigma_table[2] = sigma;
        }
        else
            return EXIT_FAILURE;
    }

    return EXIT_SUCCESS;
}
